import inspect
import json
from jwcrypto import jwk, jwt
from clearskies.handlers.exceptions import ClientError, NotFound
from clearskies.column_types import Audit, String, DateTime
from .password_login import PasswordLogin


class PasswordLessLinkLogin(PasswordLogin):
    _configuration_defaults = {
        "user_model_class": "",
        "key_column_name": "",
        "key_expiration_column_name": "",
        "username_column_name": "",
        "key_source": "query_parameters",
        "key_source_key_name": None,
        "jwt_lifetime_seconds": 86400,
        "issuer": "",
        "audience": "",
        "path_to_private_keys": "",
        "path_to_public_keys": "",
        "key_cache_duration": 7200,
        "claims_callable": None,
        "claims_column_names": None,
        "login_check_callables": [],
        "audit": True,
        "audit_column_name": "audit",
        "audit_action_name_successful_login": "login",
        "audit_action_name_failed_login": "failed_login",
        "audit_overrides": {},
        "users": None,
    }

    _required_configurations = [
        "user_model_class",
        "key_column_name",
        "key_expiration_column_name",
        "username_column_name",
        "issuer",
        "audience",
        "path_to_private_keys",
        "path_to_public_keys",
    ]

    def __init__(self, di, secrets, datetime):
        super().__init__(di, secrets, datetime)
        self._columns = None

    def _my_configuration_checks(self, configuration):
        error_prefix = "Invalid configuration for handler " + self.__class__.__name__ + ":"
        self._check_required_configuration(configuration, error_prefix)
        self._check_user_model_class_configuration(configuration, error_prefix)

        user_model_class = configuration.get("user_model_class")
        user_model = self._di.build(user_model_class)
        self._columns = user_model.columns()
        self._check_claims_configuration(configuration, error_prefix)
        self._check_input_error_callable_configuration(configuration, error_prefix)
        self._check_audit_configuration(configuration, error_prefix)
        self._check_login_check_callables(configuration, error_prefix)

        key_source = configuration.get("key_source")
        if key_source and key_source not in ["query_parameters", "json_body"]:
            raise ValueError(
                f"{error_prefix} invalid value for key_source: '{key_source}'.  It must be one of 'query_parameters' or 'json_body'"
            )

        expected_types = {
            "key_column_name": String,
            "key_expiration_column_name": DateTime,
            "username_column_name": String,
        }
        for config_name in ["key_column_name", "key_expiration_column_name", "username_column_name"]:
            column_name = configuration.get(config_name)
            if column_name not in self._columns:
                raise ValueError(
                    f"{error_prefix} the provided {config_name}, '{column_name}', does not exist in the user model '{user_model_class.__name__}'"
                )
            column = self._columns[column_name]
            if not isinstance(column, expected_types[config_name]):
                name = expected_types[config_name].__name__
                raise ValueError(
                    f"{error_prefix} config {config_name} should be a clearskies column of type {name}, but is not."
                )

    def handle(self, input_output):
        login_key = self.get_login_key(input_output)
        if not login_key:
            return self.error(input_output, "Missing login key.", 404)
        if not isinstance(login_key, str):
            return self.error(input_output, "Login key  was not a string.", 404)

        key_column_name = self.configuration("key_column_name")
        users = self.users
        user = users.find(f"{key_column_name}={login_key}")
        if not user.exists:
            return self.error(input_output, "No matching login session found.", 404)
        username_column_name = self.configuration("username_column_name")
        audit_overrides = self.configuration("audit_overrides")
        audit_extra_data_unmapped = {
            "username": user.get(self.configuration("username_column_name")),
            "user_id": user.get(user.id_column_name),
        }
        audit_extra_data = {}
        for key, value in audit_overrides.items():
            audit_extra_data[value] = audit_extra_data_unmapped[key]

        key_expiration_column_name = self.configuration("key_expiration_column_name")
        expiration = user.get(key_expiration_column_name)
        if not expiration.tzinfo:
            expiration.replace(tzinfo=self._datetime.timezone.utc)
        if not expiration:
            self.audit(
                user,
                self.configuration("audit_action_name_failed_login"),
                data={
                    "reason": "Login refused: missing expiration date for login key.",
                    **audit_extra_data,
                },
            )
            return self.error(input_output, "No matching login session found.", 404)
        if expiration < self._datetime.datetime.now(self._datetime.timezone.utc):
            self.audit(
                user,
                self.configuration("audit_action_name_failed_login"),
                data={
                    "reason": "Attempt to login with single-use key failed due to expired key.",
                    **audit_extra_data,
                },
            )
            return self.error(input_output, "No matching login session found.", 404)

        # developer-defined checks
        login_check_callables = self.configuration("login_check_callables")
        if login_check_callables:
            for login_check_callable in login_check_callables:
                response = self._di.call_function(
                    login_check_callable,
                    user=user,
                    input_output=input_output,
                    **input_output.routing_data(),
                    **input_output.context_specifics(),
                )
                if response:
                    self.audit(
                        user,
                        self.configuration("audit_action_name_failed_login"),
                        data={
                            "reason": response,
                            **audit_extra_data,
                        },
                    )
                    return self.error(input_output, "No matching login session found.", 404)

        signing_key = self.get_youngest_private_key(self.configuration("path_to_private_keys"), as_json=False)
        jwt_claims = self.get_jwt_claims(user)
        token = jwt.JWT(header={"alg": "RS256", "typ": "JWT", "kid": signing_key["kid"]}, claims=jwt_claims)
        token.make_signed_token(signing_key)
        self.audit(user, self.configuration("audit_action_name_successful_login"), data=audit_extra_data)
        user.save(
            {
                key_column_name: "",
                key_expiration_column_name: self._datetime.datetime.now(self._datetime.timezone.utc),
            }
        )

        return self.respond_unstructured(
            input_output,
            {
                "token": token.serialize(),
                "expires_at": jwt_claims["exp"],
            },
            200,
        )

    def get_login_key(self, input_output):
        key_source = self.configuration("key_source")
        key_source_key_name = self.configuration("key_source_key_name")
        if not key_source_key_name:
            key_source_key_name = self.configuration("key_column_name")

        if key_source == "query_parameters":
            return input_output.get_query_parameter(key_source_key_name)
        request_data = input_output.json_body()
        return request_data.get(key_source_key_name)



from clearskies.column_types import BelongsTo


class TenantId(BelongsTo):
    required_configs = [
        "parent_models_class",
        "source",
        "source_key_name",
    ]

    my_configs = [
        "model_column_name",
        "readable_parent_columns",
        "join_type",
    ]

    def __init__(self, di):
        super().__init__(di)

    @property
    def is_writeable(self):
        return False

    def _check_configuration(self, configuration):
        super()._check_configuration(configuration)
        error_prefix = f"Error for column '{self.name}' in model '{self.model_class.__name__}':"
        if configuration.get("source") not in ["authorization_data", "routing_data"]:
            raise ValueError(
                f"{error_prefix} 'source' must be either 'authorization_data' or 'routing_data', but was something else."
            )

    def _get_tenant_id(self):
        input_output = self.di.build("input_output", cache=True)
        source = self.config("source")
        source_key_name = self.config("source_key_name")
        data = input_output.get_authorization_data() if source == "authorization_data" else input_output.routing_data()
        tenant_id = data.get(source_key_name)
        if not tenant_id:
            raise ValueError(f"I couldn't find the tenant id in the {source} under the key {source_key_name}")
        return tenant_id

    def pre_save(self, data, model):
        # we need to provide the tenant id during a create operation
        if not model.exists:
            data[self.name] = self._get_tenant_id()
        return data

    def where_for_request(self, models, routing_data, authorization_data, input_output):
        """
        A hook to automatically apply filtering whenever the column makes an appearance in a get/update/list/search handler.
        """
        table_name = models.table_name()
        return models.where(f"{table_name}.{self.name}=" + self._get_tenant_id())



import secrets
import inspect
from clearskies.handlers.exceptions import InputError
from clearskies.handlers import Update
from clearskies.column_types import Audit


class PasswordReset(Update):
    _configuration_defaults = {
        "user_model_class": "",
        "password_column_name": "password",
        "reset_key_column_name": "reset_key",
        "reset_expiration_column_name": "reset_key_expiration",
        "reset_key_source": "routing_data",
        "reset_key_source_key_name": "reset_key",
        "model_class": "",
        "column_overrides": None,
        "columns": None,
        "writeable_columns": None,
        "readable_columns": None,
        "where": [],
        "input_error_callable": None,
        "audit": True,
        "audit_column_name": None,
        "audit_action_name_successful_reset": "password_reset",
    }

    _required_configurations = [
        "user_model_class",
    ]

    def __init__(self, di, datetime):
        super().__init__(di)
        self._columns = None
        self._datetime = datetime

    def _check_configuration(self, configuration):
        user_model_class = configuration.get("user_model_class")
        super()._check_configuration(
            {
                **{
                    "model_class": user_model_class,
                    "writeable_columns": [
                        configuration.get(
                            "password_column_name", self._configuration_defaults.get("password_column_name")
                        )
                    ],
                },
                **configuration,
            }
        )
        error_prefix = "Configuration error for %s:" % (self.__class__.__name__)
        if "reset_key_source" in configuration and configuration["reset_key_source"] not in ["routing_data"]:
            raise ValueError(f"{error_prefix} 'reset_key_source' must be 'routing_data'")

        columns_to_check = [
            "password_column_name",
            "reset_key_column_name",
            "reset_expiration_column_name",
        ]
        for config_name in columns_to_check:
            is_default = config_name not in configuration
            column_name = configuration.get(config_name, self._configuration_defaults.get(config_name))
            if column_name not in self._columns:
                if is_default:
                    raise ValueError(
                        f"{error_prefix} the configuration setting, '{config_name}' is not set and the default column name, '{column_name}', does not exist in the user model '{user_model_class.__name__}'"
                    )
                else:
                    raise ValueError(
                        f"{error_prefix} the provided column name for {config_name}, '{column_name}', does not exist in the user model '{user_model_class.__name__}'"
                    )

        if configuration.get("audit"):
            audit_column_name = configuration.get("audit_column_name")
            if audit_column_name not in self._columns:
                raise ValueError(
                    f"{error_prefix} 'audit_column_name' is '{audit_column_name}' but this column does not exist in the user model class, '{user_model_class.__name__}'"
                )
            if not isinstance(self._columns[audit_column_name], Audit):
                raise ValueError(
                    f"{error_prefix} 'audit_column_name' is '{audit_column_name}' but this column is not an audit column for the user model class, '{user_model_class.__name__}'"
                )

    def _get_audit_column(self, columns):
        audit_column = None
        for column in columns.values():
            if not isinstance(column, Audit):
                continue
            return column
        return None

    def apply_default_configuration(self, configuration):
        if not configuration.get("audit_column_name") and ("audit" not in configuration or configuration["audit"]):
            configuration["audit_column_name"] = self._get_audit_column(self._columns).name
        return super().apply_default_configuration(
            {
                **{
                    "model_class": configuration.get("user_model_class"),
                    "writeable_columns": [
                        configuration.get(
                            "password_column_name", self._configuration_defaults.get("password_column_name")
                        )
                    ],
                },
                **configuration,
            }
        )

    @property
    def users(self):
        return self._di.build(self.configuration("user_model_class"), cache=True)

    def get_model_id(self, input_output, input_data):
        routing_data = input_output.routing_data()
        reset_key_column_name = self.configuration("reset_key_column_name")
        reset_key_source_key_name = self.configuration("reset_key_source_key_name")
        reset_expiration_column_name = self.configuration("reset_expiration_column_name")
        if not routing_data.get(reset_key_source_key_name):
            raise ValueError(
                "Error with PasswordReset handler: the reset key wasn't found in the routing data, which usually means I'm misconfigured"
            )

        user = self.users.find(f"{reset_key_column_name}=" + routing_data.get(reset_key_source_key_name))
        if not user.exists:
            return None

        # make sure it hasn't expired
        expiration = user.get(reset_expiration_column_name)
        now = (
            self._datetime.datetime.now(self._datetime.timezone.utc)
            if expiration.tzinfo
            else self._datetime.datetime.now()
        )
        if expiration < now:
            return None

        return user.get(self.users.id_column_name)

    def documentation(self):
        return self._documentation(
            description="Reset the user's password",
            response_description=f"Success or Failure",
            include_id_in_path=False,
        )



from clearskies import column_types
from clearskies.functional import validations


class MultiTenantConfiguration(column_types.Column):
    my_configs = [
        "config_model_class",
        "config_model_class_user_id_column_name",
        "authorization_data_user_id_column_name",
        "config_model_class_tenant_id_column_name",
        "authorization_data_tenant_id_column_name",
        "writeable_column_names",
        "readable_column_names",
    ]

    def __init__(self, di):
        super().__init__(di)

    @property
    def is_readable(self):
        return False

    def _check_configuration(self, configuration):
        super()._check_configuration(configuration)
        # pull out the columns from the config model class and the given model class
        # Validate that there are no overlapping writable columns between the config model class and the
        # given model class
        error_prefix = f"Error for column '{self.name}' in model '{self.model_class.__name__}':"
        config_model_class = configuration.get("config_model_class")
        if not config_model_class:
            raise ValueError(
                f"{error_prefix} you must provide 'config_model_class' when using the multi_tenant_configuration column type"
            )
        if not validations.is_model_class(config_model_class):
            raise ValueError(
                f"{error_prefix} the provided value for 'config_model_class' must be a clearskies model class, but instead I got: "
                + config_model_class.__class__.__name__
            )
        self.columns = self.di.build(self.model_class, cache=True).columns()
        self.config_models = self.di.build(config_model_class)
        self.config_model_columns = self.config_models.columns()
        self._check_column_names(
            self.configuration, "writeable_column_names", self.columns, self.config_model_columns, writeable=True
        )
        self._check_column_names(
            self.configuration, "readable_column_names", self.columns, self.config_model_columns, writeable=False
        )
        self._check_config_model_column_name(
            configuration, "config_model_class_user_id_column_name", config_model_class, self.config_model_columns
        )
        if configuration.get("input_requirements"):
            raise ValueError(
                f"{error_prefix} input requirements are not allowed for the 'multi_tenant_configuration' column class"
            )

    def _check_column_names(configuration, key, model_columns, config_columns, writeable=False):
        config_model_class_name = configuration.get("config_model_class").__name__
        if not configuration.get(key):
            raise ValueError(
                f"{error_prefix} you must provide '{key}', which should be a list of column names from the config model class"
            )
        column_names = configuration.get(key)
        if not isinstance(column_names, list):
            raise ValueError(
                f"{error_prefix} '{key}' should be a list of column names, but instead of a list it is a "
                + column_names.__class__.__name__
            )
        for index, column_name in enumerate(column_names):
            if not isinstance(column_name, str):
                raise ValueError(
                    f"{error_prefix} '{key}' should be a list of column names, but entry #{index+1} is not a string.  Instead, it is a "
                    + column_name.__class__.__name__
                )
            if column_name not in config_columns:
                from_provides = False
                if not writeable:
                    for model_column in model_columns:
                        if model_column.can_provide(column_name):
                            from_provides = True
                if not from_provides:
                    raise ValueError(
                        f"{error_prefix} '{key}' should be a list of column names from the config model class, '{config_model_class_name}', but entry #{index+1} specifies a column name, '{column_name}' that does not exist in the config model class"
                    )
            if column_name in model_columns:
                raise ValueError(
                    f"{error_prefix} entry #{index+1} in '{key}' specifies a column in the config model class, '{column_name}', that already exists in the model class.  This is not allowed."
                )
            if writeable and not config_columns[column_name].is_writeable:
                raise ValueError(
                    f"{error_prefix} entry #{index+1} in '{key}' specifies a writeable column in the config model class, '{column_name}', but according to the config model class this column is not writeable."
                )
            if not writeable and not config_columns[column_name].is_readable:
                raise ValueError(
                    f"{error_prefix} entry #{index+1} in '{key}' specifies a readable column in the config model class, '{column_name}', but according to the config model class this column is not readable."
                )

    def _check_config_model_column_name(configuration, config_name, config_model_class, config_model_columns):
        config_model_column_name = configuration.get(config_name)
        if not config_model_column_name:
            raise ValueError(
                f"{error_prefix} you must provide '{config_name}' when using the multi_tenant_configuration column type.  It should be the name of a column in the config model class."
            )
        if not isinstance(config_model_column_name, str):
            raise ValueError(
                f"{error_prefix} the value specified in '{config_model_column_name}' should be a string, but instead it is a "
                + config_model_column_name.__class__.__name__
            )
        if config_model_column_name not in config_model_columns:
            raise ValueError(
                f"{error_prefix} the column name specified in '{config_name}', '{config_model_column_name}', does not exist in the config model class, '{config_model_class.__name__}'"
            )

    def post_save(self, data, model, id):
        # gather our data, find the config model class, and save.
        # also, remove the column data before returning so the other columns
        # won't trigger their post-saves
        config_model_data = {}
        for column_name in self.config("writeable_column_names"):
            if column_name not in data:
                continue
            config_model_data[column_name] = data[column_name]
            del data[column_name]

        if config_model_data:
            # model.get() returns None if the model doesn't exist, so we have to split up create and update
            # we kinda have to anyway, because with create we need to also provide the user id and tenant id.
            config_model = model.get(self.name) if model.exists else self.config_models.blank()
            if not config_model.exists:
                config_model_data = {
                    **config_model_data,
                    **self._get_record_selector_data(),
                }
            config_model.save(config_model_data)

        return data

    def additional_write_columns(self, is_create=False):
        extra_columns = super().additional_write_columns(is_create=is_create)

        # we need to have a "placeholder" for all of our writeable columns in the "root" list of columns.
        # This way, they will appear as "normal" columns in the schema, and the input checking won't complain
        # about columns that don't exist.  We'll want to make them completely new using the same "base" class.
        # this will make autodocs mostly work out.  For instance, BelongsTo will get mapped to String.
        base_classes = [
            column_types.Boolean,
            column_types.DateTime,
            column_types.Float,
            column_types.Integer,
            column_types.JSON,
            column_types.String,
        ]
        for column_name in self.config("writeable_column_names"):
            column_to_clone = self.config_model_columns[column_name]
            # again, it's not obvious but this is doing something important.  Derived classes get
            # re-built using a few of our key base column classes.
            for base_class in base_classes:
                if isinstance(column_to_clone, base_class):
                    new_class = base_class
            if not new_class:
                raise ValueError(
                    "Apparently I just wasn't designed to handle classes of column type "
                    + column_to_clone.__class__.__name__
                )
            new_column = self.di.build(new_class)
            new_column.configure()
            extra_columns.update(column_name, new_column)
        return extra_columns

    def can_provide(self, column_name):
        return column_name == self.name or column_name in self.configuration("readable_column_names")

    def provide(self, data, column_name):
        if column_name == self.name:
            models = self.config_models
            for column_name, value in self._get_record_selector_data():
                models = models.where(f"{column_name}={value}")
            return models.first()

        return self.provide(data, column_name).get(column_name)

    def to_json(self, model):
        json_data = {}
        config_model = model.get(self.name)
        for column_name in self.readable_column_names:
            json_data = {
                **json_data,
                **self.config_model_columns[column_name].to_json(config_model),
            }
        return json_data

    def _get_record_selector_data(self):
        authorization_data = self.di.build("input_output", cache=True).get_authorization_data()
        user_id_column_name = self.config("config_model_class_user_id_column_name")
        tenant_id_column_name = self.config("config_model_class_tenant_id_column_name")
        user_id = authorization_data.get(self.config("authorization_data_user_id_column_name"))
        tenant_id = authorization_data.get(self.config("authorization_data_tenant_id_column_name"))
        if not user_id:
            raise ValueError(
                "I was asked to fetch the multi tenant config for a given user, but I didn't find any data in the authorization data under "
                + self.config("authorization_data_user_id_column_name")
            )
        if not tenant_id:
            raise ValueError(
                "I was asked to fetch the multi tenant config for a given user, but I didn't find any data in the authorization data under "
                + self.config("authorization_data_tenant_id_column_name")
            )
        return {
            user_id_column_name: user_id,
            authorization_data_tenant_id_column_name: tenant_id,
        }

    def to_backend(self, data):
        data = {**data}
        for column_name in self.config("writeable_column_names"):
            if column_name in data:
                del data[column_name]
        return data

    def input_errors(self, model, data):
        config_model_data = {}
        for column_name in self.config("writeable_column_names"):
            if column_name in data:
                config_model_data[column_name] = data[column_name]

        # yeah, two for loops over the same thing, but we want
        # to collect all the data before calling check_input, because
        # normally clearskies passes around all the save data during input
        # validation.  There can be some edge cases where things might
        # break if we only send along data for a single column.
        errors = {}
        for column_name in self.config("writeable_column_names"):
            errors = {
                **errors,
                **self.config_model_columns[column_name].input_errors(model, data),
            }

        # I've skipped the part where I check input requirements for *this* column.
        # In general, that just doesn't make any sense, and so we throw an exception
        # in the configuration check stage if the developer tries to set input requirements
        # for this column
        return errors



from .create_key import CreateKey
from .delete_key import DeleteKey
from .delete_oldest_key import DeleteOldestKey
from .jwks import Jwks
from .key_base import KeyBase
from .key_base_test_helper import KeyBaseTestHelper
from .list_keys import ListKeys
from .password_less_link_login import PasswordLessLinkLogin
from .password_login import PasswordLogin
from .password_reset import PasswordReset
from .password_reset_request import PasswordResetRequest
from .profile import Profile
from .switch_tenant import SwitchTenant

# from .password_less_email_request_login import PasswordLessEmailRequestLogin

__all__ = [
    "CreateKey",
    "DeleteKey",
    "DeleteOldestKey",
    "ListKeys",
    "KeyBase",
    "KeyBaseTestHelper",
    "Jwks",
    "PasswordLessLinkLogin",
    "PasswordLogin",
    "PasswordReset",
    "PasswordResetRequest",
    "Profile",
    "SwitchTenant",
]
